#!/usr/bin/env python3
"""
This module provides a material class that can do thermochemical calculations.
"""

from auxi.core.objects import NamedObject


__version__ = '0.3.4'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Johan Zietsman'
__credits__ = ['Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Project(NamedObject):
    """
    A modelling project.

    :param name: A name for the project.
    :param description: the project's description
    """

    def __init__(self, name, description=None):
        super().__init__(name, description)


class Variable(NamedObject):
    """
    A variable used in a modelling project.

    :param name: A name for the variable.
    :param description: the variable's description
    """

    def __init__(self, name, value, units=None, category=None,
                 description=None):
        super(Variable, self).__init__(name, description)
        self.category = category
        self.value = value
        self.units = units

    def __call__(self):
        return self.value


class VariableGroup(NamedObject):
    """
    A variable group used in a modelling project.

    :param name: A name for the group.
    :param description: the group's description
    """

    def __init__(self, name, value, units=None, category=None,
                 description=None):
        super(Variable, self).__init__(name, description)

    def __call__(self):
        return self.value


if __name__ == '__main__':
    import unittest
    from core_test import *
    unittest.main()



#!/usr/bin/env python3
"""
This module provides a material class that can do thermochemical calculations.
"""

from auxi.core.objects import NamedObject


__version__ = '0.3.4'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Johan Zietsman'
__credits__ = ['Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Model(NamedObject):
    """
    A process model.

    :param name: A name for the model.
    :param description: the model's description
    """

    def __init__(self, name, description=None):
        super().__init__(name, description)

        self.parameters = type(self).Parameters()
        self.input_variables = type(self).InputVariables()
        self.output_variables = type(self).OutputVariables()

    def run(self, streams):
        """
        Run the model.

        :param streams: A dictionary of streams in the model.
        :returns: The streams dictionary with the output streams calculated by
        this model.

        If the input streams to required by the model is not found in the
        dictionary, the model will create the streams by itself based on its
        parameters and input variables.
        """
        pass

    class Parameters(object):
        pass

    class InputVariables(object):
        pass

    class OutputVariables(object):
        pass


class SteadyStateModel(Model):
    """
    A steady state process model.

    :param name: A name for the model.
    :param description: the model's description
    """

    def __init__(self, name, description=None):
        super().__init__(name, description)


class DynamicModel(Model):
    """
    A dynamic process model.

    :param name: A name for the model.
    :param description: the model's description
    """

    def __init__(self, name, description=None):
        super().__init__(name, description)


if __name__ == '__main__':
    import unittest
    from core_test import *
    unittest.main()



#!/usr/bin/env python3
"""
This module provides classes and functions for doing thermochemical
calculations.
"""

import os
import sys
import glob
import math
import warnings
import json

from auxi.core.objects import Object, NamedObject
from auxi.core.helpers import get_path_relative_to_module as get_path
from auxi.tools.chemistry.stoichiometry import molar_mass as mm
from auxi.tools.physicalconstants import R


__version__ = '0.3.1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class CpRecord(Object):
    """
    A heat capacity (Cp) equation record for a compound phase over a
    specific temperature range.

    :param dictionary: A dictionary containing the data required to initialise
      the phase.
    """

    def __init__(self, dictionary):
        self.Tmin = dictionary['Tmin']
        """[K] The minimum temperature of the range covered by this record."""
        if type(self.Tmin) is not float:
            self.Tmin = float(self.Tmin)

        self.Tmax = dictionary['Tmax']
        """[K] The maximum temperature of the range covered by this record."""
        if type(self.Tmax) is not float:
            self.Tmax = float(self.Tmax)

        self._coefficients = []
        """The coefficients of the terms in the Cp equation."""

        self._exponents = []
        """The exponents of the terms in the Cp equation."""

        for t in dictionary['Terms']:
            c = t['Coefficient']
            e = t['Exponent']
            if type(c) is not float:
                c = float(c)
            if type(e) is not float:
                e = float(e)
            self._coefficients.append(c)
            self._exponents.append(e)

    def __str__(self):
        result = '\t\tCp RECORD:' + '\n'
        result += '\t\t\tTmin: ' + str(self.Tmin) + '\n'
        result += '\t\t\tTmax: ' + str(self.Tmax) + '\n'

        for i in range(len(self._coefficients)):
            result += '\t\t\t{:.8e}'.format(self._coefficients[i]) + ' '
            result += str(self._exponents[i]) + '\n'

        return result

    def Cp(self, T):
        """
        Calculate the heat capacity of the compound phase.

        :param T: [K] temperature

        :returns: [J/mol/K] Heat capacity.
        """

        result = 0.0
        for c, e in zip(self._coefficients, self._exponents):
            result += c*T**e
        return result

    def H(self, T):
        """
        Calculate the portion of enthalpy of the compound phase covered by this
        Cp record.

        :param T: [K] temperature

        :returns: [J/mol] Enthalpy.
        """

        result = 0.0
        if T < self.Tmax:
            lT = T
        else:
            lT = self.Tmax
        Tref = self.Tmin

        for c, e in zip(self._coefficients, self._exponents):
            # Analytically integrate Cp(T).
            if e == -1.0:
                result += c * math.log(lT/Tref)
            else:
                result += c * (lT**(e+1.0) - Tref**(e+1.0)) / (e+1.0)
        return result

    def S(self, T):
        """
        Calculate the portion of entropy of the compound phase covered by this
        Cp record.

        :param T: [K] temperature

        :returns: Entropy. [J/mol/K]
        """

        result = 0.0
        if T < self.Tmax:
            lT = T
        else:
            lT = self.Tmax
        Tref = self.Tmin
        for c, e in zip(self._coefficients, self._exponents):
            # Create a modified exponent to analytically integrate Cp(T)/T
            # instead of Cp(T).
            e_modified = e - 1.0
            if e_modified == -1.0:
                result += c * math.log(lT/Tref)
            else:
                e_mod = e_modified + 1.0
                result += c * (lT**e_mod - Tref**e_mod) / e_mod
        return result


class Phase(NamedObject):
    """
    A phase of a chemical compound.

    :param dictionary: Dictionary containing the data required to initialise
      the phase.
    """

    def __init__(self, dictionary):
        self.name = dictionary['Symbol']
        """The phase's name, e.g. solid, liquid, gas, etc."""

        self.symbol = dictionary['Symbol']
        """The phase's symbol, e.g. S1 = solid 1, L = liquid, etc."""

        self.Tref = 298.15
        """[K] The reference temperature of the phase."""

        self.DHref = dictionary['DHref']
        """[J/mol] The formation enthalpy of the phase at Tref."""

        self.Sref = dictionary['Sref']
        """[J/mol/K] The standard entropy of the phase at Tref."""

        if 'magnetic' in dictionary:
            self.Tc_mag = ['magnetic']['Tc']
            """The critical temperature, which is the Curie temperature for
            ferromagnetic materials or the Neel temperature for
            antiferromagnetic materials."""

            self.beta0_mag = ['magnetic']['beta0']
            """The average magnetic moment per atom."""

            self.p_mag = ['magnetic']['p']
            """This value can be thought of as the fraction of the magnetic
            enthalpy absorbed above the critical. It depends on structure."""

        self._Cp_records = {}
        """A dictionary containing the phase's Cp records."""

        for k, v in dictionary['Cp_records'].items():
            tmax = k
            # TODO: Fix string/float conversion issue
            tmax = str(float(tmax))
            self._Cp_records[tmax] = CpRecord(v)

        self._init()

    def _init(self):
        if 'Tc_mag' in dir(self) and 'beta0_mag' in dir(self) and \
           'p_mag' in dir(self):
            self._A_mag = 79/(140*self.p_mag)
            self._B_mag = (474/497)*(1/self.p_mag - 1)
            self._D_mag = (518/1125) + (11692/15975)*(1/self.p_mag - 1)
        else:
            self.Cp_mag = self.Zero_mag
            self.H_mag = self.Zero_mag
            self.S_mag = self.Zero_mag
            self.G_mag = self.Zero_mag

    def __str__(self):
        result = '\tPHASE: ' + self.name + '\n'
        result += '\t\tName: ' + self.name + '\n'
        result += '\t\tSymbol: ' + self.symbol + '\n'
        result += '\t\tTref: ' + str(self.Tref) + '\n'
        result += '\t\tDHref: ' + str(self.DHref) + '\n'
        result += '\t\tSref: ' + str(self.Sref) + '\n'
        result += '\t\tCp record count:' + str(len(self._Cp_records)) + '\n'

        for k, v in self._Cp_records.items():
            result += str(v)

        return result

    def Zero_mag(self, T):
        """
        Return a zero value for a phase with no magnetic property data.

        :param T: [K] temperature

        :returns: Zero.
        """

        return 0.0

    def Cp(self, T):
        """
        Calculate the heat capacity of the compound phase at the specified
        temperature.

        :param T: [K] temperature

        :returns: [J/mol/K] The heat capacity of the compound phase.
        """

        # TODO: Fix str/float conversion
        for Tmax in sorted([float(TT) for TT in self._Cp_records.keys()]):
            if T < Tmax:
                return self._Cp_records[str(Tmax)].Cp(T) + self.Cp_mag(T)

        Tmax = max([float(TT) for TT in self._Cp_records.keys()])

        return self._Cp_records[str(Tmax)].Cp(Tmax) + self.Cp_mag(T)

    def Cp_mag(self, T):
        """
        Calculate the phase's magnetic contribution to heat capacity at the
        specified temperature.

        :param T: [K] temperature

        :returns: [J/mol/K] The magnetic heat capacity of the compound phase.

        Dinsdale, A. T. (1991). SGTE data for pure elements. Calphad, 15(4),
        317–425. http://doi.org/10.1016/0364-5916(91)90030-N
        """

        tau = T / self.Tc_mag

        if tau <= 1.0:
            c = (self._B_mag*(2*tau**3 + 2*tau**9/3 + 2*tau**15/5))/self._D_mag
        else:
            c = (2*tau**-5 + 2*tau**-15/3 + 2*tau**-25/5)/self._D_mag

        result = R*math.log(self.beta0_mag + 1)*c

        return result

    def H(self, T):
        """
        Calculate the enthalpy of the compound phase at the specified
        temperature.

        :param T: [K] temperature

        :returns: [J/mol] The enthalpy of the compound phase.
        """

        result = self.DHref

        for Tmax in sorted([float(TT) for TT in self._Cp_records.keys()]):
            result += self._Cp_records[str(Tmax)].H(T)
            if T <= Tmax:
                return result + self.H_mag(T)

        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max([float(TT) for TT in self._Cp_records.keys()])
        result += self.Cp(Tmax)*(T - Tmax)

        return result + self.H_mag(T)

    def H_mag(self, T):
        """
        Calculate the phase's magnetic contribution to enthalpy at the
        specified temperature.

        :param T: [K] temperature

        :returns: [J/mol] The magnetic enthalpy of the compound phase.

        Dinsdale, A. T. (1991). SGTE data for pure elements. Calphad, 15(4),
        317–425. http://doi.org/10.1016/0364-5916(91)90030-N
        """

        tau = T / self.Tc_mag

        if tau <= 1.0:
            h = (-self._A_mag/tau +
                 self._B_mag*(tau**3/2 + tau**9/15 + tau**15/40))/self._D_mag
        else:
            h = -(tau**-5/2 + tau**-15/21 + tau**-25/60)/self._D_mag

        return R*T*math.log(self.beta0_mag + 1)*h

    def S(self, T):
        """
        Calculate the entropy of the compound phase at the specified
        temperature.

        :param T: [K] temperature

        :returns: [J/mol/K] The entropy of the compound phase.
        """

        result = self.Sref

        for Tmax in sorted([float(TT) for TT in self._Cp_records.keys()]):
            result += self._Cp_records[str(Tmax)].S(T)
            if T <= Tmax:
                return result + self.S_mag(T)

        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max([float(TT) for TT in self._Cp_records.keys()])
        result += self.Cp(Tmax)*math.log(T / Tmax)

        return result + self.S_mag(T)

    def S_mag(self, T):
        """
        Calculate the phase's magnetic contribution to entropy at the
        specified temperature.

        :param T: [K] temperature

        :returns: [J/mol/K] The magnetic entropy of the compound phase.

        Dinsdale, A. T. (1991). SGTE data for pure elements. Calphad, 15(4),
        317–425. http://doi.org/10.1016/0364-5916(91)90030-N
        """

        tau = T / self.Tc_mag

        if tau <= 1.0:
            s = 1 - (self._B_mag*(2*tau**3/3 + 2*tau**9/27 + 2*tau**15/75)) / \
                self._D_mag
        else:
            s = (2*tau**-5/5 + 2*tau**-15/45 + 2*tau**-25/125)/self._D_mag

        return -R*math.log(self.beta0_mag + 1)*s

    def G(self, T):
        """Calculate the heat capacity of the compound phase at the specified
        temperature.

        :param T: [K] temperature

        :returns: [J/mol] The Gibbs free energy of the compound phase.
        """

        h = self.DHref
        s = self.Sref

        for Tmax in sorted([float(TT) for TT in self._Cp_records.keys()]):
            h = h + self._Cp_records[str(Tmax)].H(T)
            s = s + self._Cp_records[str(Tmax)].S(T)
            if T <= Tmax:
                return h - T * s + self.G_mag(T)

        # Extrapolate beyond the upper limit by using a constant heat capacity.
        Tmax = max([float(TT) for TT in self._Cp_records.keys()])
        h = h + self.Cp(Tmax)*(T - Tmax)
        s = s + self.Cp(Tmax)*math.log(T / Tmax)

        return h - T * s + self.G_mag(T)

    def G_mag(self, T):
        """
        Calculate the phase's magnetic contribution to Gibbs energy at the
        specified temperature.

        :param T: [K] temperature

        :returns: [J/mol] The magnetic Gibbs energy of the compound phase.

        Dinsdale, A. T. (1991). SGTE data for pure elements. Calphad, 15(4),
        317–425. http://doi.org/10.1016/0364-5916(91)90030-N
        """

        tau = T / self.Tc_mag

        if tau <= 1.0:
            g = 1 - (self._A_mag/tau +
                     self._B_mag*(tau**3/6 + tau**9/135 + tau**15/600)) /\
                    self._D_mag
        else:
            g = -(tau**-5/10 + tau**-15/315 + tau**-25/1500)/self._D_mag

        return R*T*math.log(self.beta0_mag + 1)*g


class Compound(Object):
    """
    Represents a chemical compound.

    :param dictionary: Dictionary containing the data required to initialise
      the compound.
    """

    def __init__(self, dictionary):
        self.formula = dictionary['Formula']
        """Chemical formula, e.g. 'Fe', 'CO2'."""

        self.molar_mass = mm(self.formula) / 1000.0
        """Molar mass. [kg/mol]"""

        self._phases = {}
        """Dictionary containing the compound's phase objects."""

        if 'Reference' in dictionary:
            self.reference = dictionary['Reference']
        else:
            self.reference = ""

        """Reference to the publisher of the thermo data."""

        for k, v in dictionary['Phases'].items():
            if not 'Symbol' in v:
                v['Symbol'] = k
            self._phases[k] = Phase(v)

#    def __str__(self):
#        result = 'COMPOUND: ' + '\n'
#        result += '\tFormula: ' + self.formula + '\n'
#
#        for k, v in self._phases.items():
#            result += str(v)
#
#        return result

    def _init(self):
        for p in self._phases:
            self._phases[p]._init()

    def get_phase_list(self):
        """
        Get a list of the compound's phases.

        :returns: List of phases.
        """

        return sorted(self._phases.keys())

    def get_reference(self):
        if self.reference in [None, ""]:
            return ""
        return get_datafile_references()[self.reference]

    def Cp(self, phase, T):
        """
        Calculate the heat capacity of a phase of the compound at a specified
        temperature.

        :param phase: A phase of the compound, e.g. 'S', 'L', 'G'.
        :param T: [K] temperature

        :returns: [J/mol/K] Heat capacity.
        """

        if phase not in self._phases:
            raise Exception("The phase '%s' was not found in compound '%s'." %
                            (phase, self.formula))

        return self._phases[phase].Cp(T)

    def H(self, phase, T):
        """
        Calculate the enthalpy of a phase of the compound at a specified
        temperature.

        :param phase: A phase of the compound, e.g. 'S', 'L', 'G'.
        :param T: [K] temperature

        :returns: [J/mol] Enthalpy.
        """

        try:
            return self._phases[phase].H(T)
        except KeyError:
            raise Exception("The phase '{}' was not found in compound '{}'."
                            .format(phase, self.formula))

    def S(self, phase, T):
        """
        Calculate the enthalpy of a phase of the compound at a specified
        temperature.

        :param phase: A phase of the compound, e.g. 'S', 'L', 'G'.
        :param T: [K] temperature

        :returns: [J/mol/K] Entropy.
        """

        try:
            return self._phases[phase].S(T)
        except KeyError:
            raise Exception("The phase '{}' was not found in compound '{}'."
                            .format(phase, self.formula))

    def G(self, phase, T):
        """
        Calculate the Gibbs free energy of a phase of the compound at a
        specified temperature.

        :param phase: A phase of the compound, e.g. 'S', 'L', 'G'.
        :param T: [K] temperature

        :returns: [J/mol] Gibbs free energy.
        """

        try:
            return self._phases[phase].G(T)
        except KeyError:
            raise Exception("The phase '{}' was not found in compound '{}'."
                            .format(phase, self.formula))


def _get_default_data_path_():
    """
    Calculate the default path in which thermochemical data is stored.

    :returns: Default path.
    """

    module_path = os.path.dirname(sys.modules[__name__].__file__)
    data_path = os.path.join(module_path, r'data/rao')
    data_path = os.path.abspath(data_path)
    return data_path


def _read_compound_from_factsage_file_(file_name):
    """
    Build a dictionary containing the factsage thermochemical data of a
    compound by reading the data from a file.

    :param file_name: Name of file to read the data from.

    :returns: Dictionary containing compound data.
    """

    with open(file_name) as f:
        lines = f.readlines()

    compound = {'Formula': lines[0].split(' ')[1]}
    # FIXME: replace with logging
    print(compound['Formula'])
    compound['Phases'] = phs = {}

    started = False
    phaseold = 'zz'
    recordold = '0'

    for line in lines:
        if started:
            if line.startswith('_'):  # line indicating end of data
                break
            line = line.replace(' 298 ', ' 298.15 ')
            line = line.replace(' - ', ' ')
            while '  ' in line:
                line = line.replace('  ', ' ')
            line = line.replace(' \n', '')
            line = line.replace('\n', '')
            strings = line.split(' ')
            if len(strings) < 2:  # empty line
                continue
            phase = strings[0]
            if phase != phaseold:  # new phase detected
                phaseold = phase
                ph = phs[phase] = {}
                ph['Symbol'] = phase
                ph['DHref'] = float(strings[2])
                ph['Sref'] = float(strings[3])
                cprecs = ph['Cp_records'] = {}
                record = strings[1]
                if record != recordold:  # new record detected
                    recordold = record
                    Tmax = float(strings[len(strings) - 1])
                    cprecs[Tmax] = {}
                    cprecs[Tmax]['Tmin'] = float(strings[len(strings) - 2])
                    cprecs[Tmax]['Tmax'] = float(strings[len(strings) - 1])
                    cprecs[Tmax]['Terms'] = []
                    t = {'Coefficient': float(strings[4]),
                         'Exponent': float(strings[5])}
                    cprecs[Tmax]['Terms'].append(t)
                    if len(strings) == 10:
                        t = {'Coefficient': float(strings[6]),
                             'Exponent': float(strings[7])}
                        cprecs[Tmax]['Terms'].append(t)
                else:  # old record detected
                    t = {'Coefficient': float(strings[2]),
                         'Exponent': float(strings[3])}
                    cprecs[Tmax]['Terms'].append(t)
                    if len(strings) == 8:
                        t = {'Coefficient': float(strings[4]),
                             'Exponent': float(strings[5])}
                        cprecs[Tmax]['Terms'].append(t)
            else:  # old phase detected
                ph = phs[phase]
                record = strings[1]
                if record != recordold:  # new record detected
                    recordold = record
                    Tmax = float(strings[len(strings) - 1])
                    cprecs = ph['Cp_records']
                    cprecs[Tmax] = {}
                    cprecs[Tmax]['Tmin'] = float(strings[len(strings) - 2])
                    cprecs[Tmax]['Tmax'] = float(strings[len(strings) - 1])
                    cprecs[Tmax]['Terms'] = []
                    t = {'Coefficient': float(strings[2]),
                         'Exponent': float(strings[3])}
                    cprecs[Tmax]['Terms'].append(t)
                    if len(strings) == 8:
                        t = {'Coefficient': float(strings[4]),
                             'Exponent': float(strings[5])}
                        cprecs[Tmax]['Terms'].append(t)
                else:  # old record detected
                    t = {'Coefficient': float(strings[2]),
                         'Exponent': float(strings[3])}
                    cprecs[Tmax]['Terms'].append(t)
                    if len(strings) == 8:
                        t = {'Coefficient': float(strings[4]),
                             'Exponent': float(strings[5])}
                        cprecs[Tmax]['Terms'].append(t)
        if line.startswith('_'):  # line indicating the start of the data
            started = True

    for name, ph in phs.items():
        cprecs = ph['Cp_records']
        first = cprecs[min(cprecs.keys())]
        first['Tmin'] = 298.15

    return compound


def _split_compound_string_(compound_string):
    """
    Split a compound's combined formula and phase into separate strings for
    the formula and phase.

    :param compound_string: Formula and phase of a chemical compound, e.g.
      'SiO2[S1]'.

    :returns: Formula of chemical compound.
    :returns: Phase of chemical compound.
    """

    formula = compound_string.replace(']', '').split('[')[0]
    phase = compound_string.replace(']', '').split('[')[1]

    return formula, phase


def _finalise_result_(compound, value, mass):
    """
    Convert the value to its final form by unit conversions and multiplying
    by mass.

    :param compound: Compound object.
    :param value: [J/mol] Value to be finalised.
    :param mass: [kg] Mass of compound.

    :returns: [kWh] Finalised value.
    """

    result = value / 3.6E6  # J/x -> kWh/x
    result = result / compound.molar_mass  # x/mol -> x/kg
    result = result * mass  # x/kg -> x

    return result


def _read_compound_from_auxi_file_(file_name):
    """
    Build a dictionary containing the auxi thermochemical data of a compound by
    reading the data from a file.

    :param file_name: Name of file to read the data from.

    :returns: Dictionary containing compound data.
    """

    with open(file_name) as f:
        return json.load(f)


def write_compound_to_auxi_file(directory, compound):
    """
    Writes a compound to an auxi file at the specified directory.

    :param dir: The directory.
    :param compound: The compound.
    """

    file_name = "Compound_" + compound.formula + ".json"
    with open(os.path.join(directory, file_name), 'w') as f:
        f.write(str(compound))


def load_data_factsage(path=''):
    """
    Load all the thermochemical data factsage files located at a path.

    :param path: Path at which the data files are located.
    """

    compounds.clear()

    if path == '':
        path = default_data_path
    if not os.path.exists(path):
        warnings.warn('The specified data file path does not exist. (%s)' % path)
        return

    files = glob.glob(os.path.join(path, 'Compound_*.txt'))

    for file in files:
        compound = Compound(_read_compound_from_factsage_file_(file))
        compounds[compound.formula] = compound


def load_data_auxi(path=''):
    """
    Load all the thermochemical data auxi files located at a path.

    :param path: Path at which the data files are located.
    """

    compounds.clear()

    if path == '':
        path = default_data_path
    if not os.path.exists(path):
        warnings.warn('The specified data file path does not exist. (%s)' % path)
        return

    files = glob.glob(os.path.join(path, 'Compound_*.json'))

    for file in files:
        compound = Compound(_read_compound_from_auxi_file_(file))
#        compound = Compound.read(file)
        compounds[compound.formula] = compound


def list_compounds():
    """
    List all compounds that are currently loaded in the thermo module, and
    their phases.
    """

    print('Compounds currently loaded:')
    for compound in sorted(compounds.keys()):
        phases = compounds[compound].get_phase_list()
        print('%s: %s' % (compound, ', '.join(phases)))
#        for phase in phases:
#            print(compound + '[' + phase + ']')


def get_datafile_references():
    """
    Retrieve all the references used by the datafiles.
    """
    with open(get_path(__file__, "data/references.json")) as f:
        return json.load(f)


def molar_mass(compound):
    """
    Determine the molar mass of a chemical compound.

    :param compound: Formula of a chemical compound, e.g. 'Fe2O3'.

    :returns: [kg/mol] Molar mass.
    """

    return mm(compound) / 1000.0


def Cp(compound_string, T, mass=1.0):
    """
    Calculate the heat capacity of the compound for the specified temperature
    and mass.

    :param compound_string: Formula and phase of chemical compound, e.g.
      'Fe2O3[S1]'.
    :param T: [°C] temperature
    :param mass: [kg]

    :returns: [kWh/K] Heat capacity.
    """

    formula, phase = _split_compound_string_(compound_string)
    TK = T + 273.15
    compound = compounds[formula]
    result = compound.Cp(phase, TK)

    return _finalise_result_(compound, result, mass)


def H(compound_string, T, mass=1.0):
    """
    Calculate the enthalpy of the compound for the specified temperature and
    mass.

    :param compound_string: Formula and phase of chemical compound, e.g.
      'Fe2O3[S1]'.
    :param T: [°C] temperature
    :param mass: [kg]

    :returns: [kWh] Enthalpy.
    """

    formula, phase = _split_compound_string_(compound_string)
    TK = T + 273.15
    compound = compounds[formula]
    result = compound.H(phase, TK)

    return _finalise_result_(compound, result, mass)


def S(compound_string, T, mass=1.0):
    """
    Calculate the entropy of the compound for the specified temperature and
    mass.

    :param compound_string: Formula and phase of chemical compound, e.g.
      'Fe2O3[S1]'.
    :param T: [°C] temperature
    :param mass: [kg]

    :returns: [kWh/K] Entropy.
    """

    formula, phase = _split_compound_string_(compound_string)
    TK = T + 273.15
    compound = compounds[formula]
    result = compound.S(phase, TK)

    return _finalise_result_(compound, result, mass)


def G(compound_string, T, mass=1.0):
    """
    Calculate the Gibbs free energy of the compound for the specified
    temperature and mass.

    :param compound_string: Formula and phase of chemical compound, e.g.
      'Fe2O3[S1]'.
    :param T: [°C] temperature
    :param mass: [kg]


    :returns: [kWh] Gibbs free energy.
    """

    formula, phase = _split_compound_string_(compound_string)
    TK = T + 273.15
    compound = compounds[formula]
    result = compound.G(phase, TK)

    return _finalise_result_(compound, result, mass)


compounds = {}
default_data_path = _get_default_data_path_()
load_data_auxi()


if __name__ == '__main__':
    import unittest
    from auxi.tools.chemistry.thermochemistry_test import ThermoFunctionTester
    unittest.main()



#!/usr/bin/env python3
"""
This module provides a number of functions for doing stoichiometry
calculations.
"""

import collections
import functools

import parsimonious

from auxi.core.objects import Object


__version__ = '0.3.4'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Element(Object):
    """
    An element in the periodic table.

    :param period: Period to which the element belongs.
    :param group: Group to which the element belongs.
    :param atomic_number: Number of protons in the element's nucleus.
    :param symbol: Element's symbol.
    :param molar_mass: [kg/kmol] Element's standard atomic mass.
    """

    # TODO: Add tests.
    # TODO: Implement validate method.

    def __init__(self, period, group, atomic_number, symbol, molar_mass):
        self.period = period
        self.group = group
        self.atomic_number = atomic_number
        self.symbol = symbol
        self.molar_mass = molar_mass
        self._validate_()

    def count(self):
        return {self.symbol: 1}

    def _validate_(self):
        pass

    def __repr__(self):
        return "Element('{}')".format(self.symbol)


def count_with_multiplier(groups, multiplier):
    """ Update group counts with multiplier

    This is for handling atom counts on groups like (OH)2

    :param groups: iterable of Group/Element
    :param multiplier: the number to multiply by

    """
    counts = collections.defaultdict(float)
    for group in groups:
        for element, count in group.count().items():
            counts[element] += count*multiplier
    return counts


class Group:
    """ Represent a part of a compound formula

    :param group: iterable of Group/Element parts
    :param multiplier: multiplier in subscript or prefix of group
    :param dotted: True of the group is like '.2H2O'

    """

    def __init__(self, group, multiplier=1, dotted=False):
        self.group = group
        self.multiplier = multiplier
        self.dotted = dotted

    def count(self):
        return count_with_multiplier(self.group, self.multiplier)

    def __repr__(self):
        return "Group({}, multiplier={}, dotted={})".format(self.group,
                                                            self.multiplier,
                                                            self.dotted)


class Compound:
    """ Represents a full compound formula

    :param group: iterable of Group/Element
    :param dottedgroup: A Group if there is a .H2O part, None otherwise
    :param phase: The phase if there is a [phase] part, None otherwise

    """

    def __init__(self, group, dottedgroup=None, phase=None):
        self.group = [group]
        if dottedgroup:
            self.group.append(dottedgroup)
        self.phase = phase

    def count(self):
        return count_with_multiplier(self.group, multiplier=1)

    def molar_mass(self):
        return sum(_element_dictionary_[element].molar_mass*count
                   for element, count in self.count().items())

    def __repr__(self):
        return "Compound({}, {})".format(self.group, self.phase)


class CompoundVisitor(parsimonious.NodeVisitor):
    """ Visitor which takes parsed tree to useful groups.

        For parallels, check the grammar.
    """
    def visit_compound(self, _, compound):
        (group, dottedgroup, phase) = compound
        return Compound(group, dottedgroup, phase)

    def visit_group(self, _, group):
        return Group(group)

    def visit_phase(self, node, _):
        (_, string, _) = node
        return string.text

    def visit_dottedgroup(self, _, dottedgroup):
        (_, number, group) = dottedgroup
        if not number:
            number = 1
        return Group([group], number, dotted=True)

    def visit_subscriptedgroup(self, _, subscriptedgroup):
        (_, group, _, number) = subscriptedgroup
        if not number:
            number = 1
        return Group([group], number)

    def visit_subscriptedelement(self, _, subscriptedelement):
        (element, number) = subscriptedelement
        return Group([element], number)

    def visit_element(self, node, _):
        return _element_dictionary_[node.text]

    def visit_number(self, node, _):
        return int(node.text)

    def generic_visit(self, node, other):
        try:
            return other[0]
        except IndexError:
            return other


grammar = parsimonious.grammar.Grammar(
    """
    compound = group dottedgroup? phase?
    group = (subscriptedgroup / subscriptedelement / element)+
    phase = "[" string "]"
    dottedgroup = "." number? group
    subscriptedgroup = "(" group ")" number?
    subscriptedelement = element number
    element = ~r"[A-Z][a-z]*"
    string = ~r"[A-Za-z0-9]+"
    number = ~r"[0-9]+"
    """)


@functools.lru_cache()
def parse_compound(string):
    visitor = CompoundVisitor()
    parsed_tree = grammar.parse(string)
    return visitor.visit(parsed_tree)


def amount(compound, mass):
    """
    Calculate the number of moles in the specified mass of a chemical compound.

    :param compound: Formula and phase of a compound, e.g. 'Fe2O3[S1]'. The
      phase may be omitted.
    :param mass: [kg]

    :returns: Amount. [kmol]
    """

    return mass / molar_mass(compound)


def amounts(masses):
    """
    Calculate the amounts from the specified compound masses.

    :param masses: [kg] dictionary, e.g. {'SiO2': 3.0, 'FeO': 1.5}

    :returns: [kmol] dictionary
    """

    return {compound: amount(compound, masses[compound])
            for compound in masses.keys()}


def amount_fractions(masses):
    """
    Calculate the mole fractions from the specified compound masses.

    :param masses: [kg] dictionary, e.g. {'SiO2': 3.0, 'FeO': 1.5}

    :returns: [mole fractions] dictionary
    """

    n = amounts(masses)
    n_total = sum(n.values())
    return {compound: n[compound]/n_total for compound in n.keys()}


def mass(compound, amount):
    """
    Calculate the mass of the specified amount of a chemical compound.

    :param compound: Formula and phase of a compound, e.g. 'Fe2O3[S1]'. The
      phase may be omitted.
    :param amount: [kmol]

    :returns: Mass. [kg]
    """

    return amount * molar_mass(compound)


def masses(amounts):
    """
    Calculate the masses from the specified compound amounts.

    :param masses: [kmol] dictionary, e.g. {'SiO2': 3.0, 'FeO': 1.5}

    :returns: [kg] dictionary
    """

    return {compound: mass(compound, amounts[compound])
            for compound in amounts.keys()}


def mass_fractions(amounts):
    """
    Calculate the mole fractions from the specified compound masses.

    :param masses: [kg] dictionary, e.g. {'SiO2': 3.0, 'FeO': 1.5}

    :returns: [mass fractions] dictionary
    """

    m = masses(amounts)
    m_total = sum(m.values())
    return {compound: m[compound]/m_total for compound in m.keys()}


def convert_compound(mass, source, target, element):
    """
    Convert the specified mass of the source compound to the target using
    element as basis.

    :param mass: Mass of from_compound. [kg]
    :param source: Formula and phase of the original compound, e.g.
      'Fe2O3[S1]'.
    :param target: Formula and phase of the target compound, e.g. 'Fe[S1]'.
    :param element: Element to use as basis for the conversion, e.g. 'Fe' or
      'O'.

    :returns: Mass of target. [kg]
    """

    # Perform the conversion.
    target_mass_fraction = element_mass_fraction(target, element)
    if target_mass_fraction == 0.0:
        # If target_formula does not contain element, just return 0.0.
        return 0.0
    else:
        source_mass_fraction = element_mass_fraction(source, element)
        return mass * source_mass_fraction / target_mass_fraction


def element_mass_fraction(compound, element):
    """
    Determine the mass fraction of an element in a chemical compound.

    :param compound: Formula of the chemical compound, 'FeCr2O4'.
    :param element: Element, e.g. 'Cr'.

    :returns: Element mass fraction.
    """

    coeff = stoichiometry_coefficient(compound, element)

    if coeff == 0.0:
        return 0.0

    formula_mass = molar_mass(compound)
    element_mass = molar_mass(element)
    return coeff * element_mass / formula_mass


def element_mass_fractions(compound, elements):
    """
    Determine the mass fractions of a list of elements in a chemical compound.

    :param compound: Formula and phase of a chemical compound, e.g.
      'Fe2O3[S1]'.
    :param elements: List of elements, ['Si', 'O', 'Fe'].

    :returns: Mass fractions.
    """

    return [element_mass_fraction(compound, element)
            for element in elements]


def elements(compounds):
    """
    Determine the set of elements present in a list of chemical compounds.

    The list of elements is sorted alphabetically.

    :param compounds: List of compound formulas and phases, e.g.
      ['Fe2O3[S1]', 'Al2O3[S1]'].

    :returns: List of elements.
    """

    elementlist = [parse_compound(compound).count().keys()
                   for compound in compounds]
    return set().union(*elementlist)


def molar_mass(compound=''):
    """Determine the molar mass of a chemical compound.

    The molar mass is usually the mass of one mole of the substance, but here
    it is the mass of 1000 moles, since the mass unit used in auxi is kg.

    :param compound: Formula of a chemical compound, e.g. 'Fe2O3'.

    :returns: Molar mass. [kg/kmol]
    """

    result = 0.0
    if compound is None or len(compound) == 0:
        return result

    compound = compound.strip()

    parsed = parse_compound(compound)

    return parsed.molar_mass()


def stoichiometry_coefficient(compound, element):
    """
    Determine the stoichiometry coefficient of an element in a chemical
    compound.

    :param compound: Formula of a chemical compound, e.g. 'SiO2'.
    :param element:  Element, e.g. 'Si'.

    :returns: Stoichiometry coefficient.
    """

    stoichiometry = parse_compound(compound.strip()).count()

    return stoichiometry[element]


def stoichiometry_coefficients(compound, elements):
    """
    Determine the stoichiometry coefficients of the specified elements in
    the specified chemical compound.

    :param compound: Formula of a chemical compound, e.g. 'SiO2'.
    :param elements: List of elements, e.g. ['Si', 'O', 'C'].

    :returns: List of stoichiometry coefficients.
    """

    stoichiometry = parse_compound(compound.strip()).count()

    return [stoichiometry[element] for element in elements]


# Initialise the module.
# Create all the elements of the periodic table and add them to the
# element dictionary.

_element_dictionary_ = {
    # period 1
    'H': Element(1, 1, 1, 'H', 1.00794),
    'He': Element(1, 18, 2, 'He', 4.002602),

    # period 2
    'Li': Element(2, 1, 3, 'Li', 6.941),
    'Be': Element(2, 2, 4, 'Be', 9.012182),
    'B': Element(2, 13, 5, 'B', 10.811),
    'C': Element(2, 14, 6, 'C', 12.0107),
    'N': Element(2, 15, 7, 'N', 14.00674),
    'O': Element(2, 16, 8, 'O', 15.9994),
    'F': Element(2, 17, 9, 'F', 18.9984032),
    'Ne': Element(2, 18, 10, 'Ne', 20.1797),

    # period 3
    'Na': Element(3, 1, 11, 'Na', 22.98977),
    'Mg': Element(3, 2, 12, 'Mg', 24.305),
    'Al': Element(3, 13, 13, 'Al', 26.981538),
    'Si': Element(3, 14, 14, 'Si', 28.0855),
    'P': Element(3, 15, 15, 'P', 30.973762),
    'S': Element(3, 16, 16, 'S', 32.066),
    'Cl': Element(3, 17, 17, 'Cl', 35.4527),
    'Ar': Element(3, 18, 18, 'Ar', 39.948),

    # period 4
    'K': Element(4, 1, 19, 'K', 39.0983),
    'Ca': Element(4, 2, 20, 'Ca', 40.078),
    'Sc': Element(4, 3, 21, 'Sc', 44.95591),
    'Ti': Element(4, 4, 22, 'Ti', 47.867),
    'V': Element(4, 5, 23, 'V', 50.9415),
    'Cr': Element(4, 6, 24, 'Cr', 51.9961),
    'Mn': Element(4, 7, 25, 'Mn', 54.938049),
    'Fe': Element(4, 8, 26, 'Fe', 55.845),
    'Co': Element(4, 9, 27, 'Co', 58.9332),
    'Ni': Element(4, 10, 28, 'Ni', 58.6934),
    'Cu': Element(4, 11, 29, 'Cu', 63.546),
    'Zn': Element(4, 12, 30, 'Zn', 65.39),
    'Ga': Element(4, 13, 31, 'Ga', 69.723),
    'Ge': Element(4, 14, 32, 'Ge', 72.61),
    'As': Element(4, 15, 33, 'As', 74.9216),
    'Se': Element(4, 16, 34, 'Se', 78.96),
    'Br': Element(4, 17, 35, 'Br', 79.904),
    'Kr': Element(4, 18, 36, 'Kr', 83.8),

    # period 5
    'Rb': Element(5, 1, 37, 'Rb', 85.4678),
    'Sr': Element(5, 2, 38, 'Sr', 87.62),
    'Y': Element(5, 3, 39, 'Y', 88.90585),
    'Zr': Element(5, 4, 40, 'Zr', 91.224),
    'Nb': Element(5, 5, 41, 'Nb', 92.90638),
    'Mo': Element(5, 6, 42, 'Mo', 95.94),
    'Tc': Element(5, 7, 43, 'Tc', 98.0),
    'Ru': Element(5, 8, 44, 'Ru', 101.07),
    'Rh': Element(5, 9, 45, 'Rh', 102.9055),
    'Pd': Element(5, 10, 46, 'Pd', 106.42),
    'Ag': Element(5, 11, 47, 'Ag', 107.8682),
    'Cd': Element(5, 12, 48, 'Cd', 112.411),
    'In': Element(5, 13, 49, 'In', 114.818),
    'Sn': Element(5, 14, 50, 'Sn', 118.71),
    'Sb': Element(5, 15, 51, 'Sb', 121.76),
    'Te': Element(5, 16, 52, 'Te', 127.6),
    'I': Element(5, 17, 53, 'I', 126.90447),
    'Xe': Element(5, 18, 54, 'Xe', 131.29),

    # period 6
    'Cs': Element(6, 1, 55, 'Cs', 132.90545),
    'Ba': Element(6, 2, 56, 'Ba', 137.327),
    'La': Element(6, 0, 57, 'La', 138.9055),
    'Ce': Element(6, 0, 58, 'Ce', 140.116),
    'Pr': Element(6, 0, 59, 'Pr', 140.90765),
    'Nd': Element(6, 0, 60, 'Nd', 144.24),
    'Pm': Element(6, 0, 61, 'Pm', 145.0),
    'Sm': Element(6, 0, 62, 'Sm', 150.36),
    'Eu': Element(6, 0, 63, 'Eu', 151.964),
    'Gd': Element(6, 0, 64, 'Gd', 157.25),
    'Tb': Element(6, 0, 65, 'Tb', 158.92534),
    'Dy': Element(6, 0, 66, 'Dy', 162.5),
    'Ho': Element(6, 0, 67, 'Ho', 164.93032),
    'Er': Element(6, 0, 68, 'Er', 167.26),
    'Tm': Element(6, 0, 69, 'Tm', 168.93421),
    'Yb': Element(6, 0, 70, 'Yb', 173.04),
    'Lu': Element(6, 0, 71, 'Lu', 174.967),
    'Hf': Element(6, 4, 72, 'Hf', 178.49),
    'Ta': Element(6, 5, 73, 'Ta', 180.9479),
    'W': Element(6, 6, 74, 'W', 183.84),
    'Re': Element(6, 7, 75, 'Re', 186.207),
    'Os': Element(6, 8, 76, 'Os', 190.23),
    'Ir': Element(6, 9, 77, 'Ir', 192.217),
    'Pt': Element(6, 10, 78, 'Pt', 195.078),
    'Au': Element(6, 11, 79, 'Au', 196.96655),
    'Hg': Element(6, 12, 80, 'Hg', 200.59),
    'Tl': Element(6, 13, 81, 'Tl', 204.3833),
    'Pb': Element(6, 14, 82, 'Pb', 207.2),
    'Bi': Element(6, 15, 83, 'Bi', 208.98038),
    'Po': Element(6, 16, 84, 'Po', 210.0),
    'At': Element(6, 17, 85, 'At', 210.0),
    'Rn': Element(6, 18, 86, 'Rn', 222.0),

    # period 7
    'Fr': Element(7, 1, 87, 'Fr', 223.0),
    'Ra': Element(7, 2, 88, 'Ra', 226.0),
    'Ac': Element(7, 0, 89, 'Ac', 227.0),
    'Th': Element(7, 0, 90, 'Th', 232.0381),
    'Pa': Element(7, 0, 91, 'Pa', 231.03588),
    'U': Element(7, 0, 92, 'U', 238.0289),
    'Np': Element(7, 0, 93, 'Np', 237.0),
    'Pu': Element(7, 0, 94, 'Pu', 244.0),
    'Am': Element(7, 0, 95, 'Am', 243.0),
    'Cm': Element(7, 0, 96, 'Cm', 247.0),
    'Bk': Element(7, 0, 97, 'Bk', 247.0),
    'Cf': Element(7, 0, 98, 'Cf', 251.0),
    'Es': Element(7, 0, 99, 'Es', 252.0),
    'Fm': Element(7, 0, 100, 'Fm', 257.0),
    'Md': Element(7, 0, 101, 'Md', 258.0),
    'No': Element(7, 0, 102, 'No', 259.0),
    'Lr': Element(7, 0, 103, 'Lr', 262.0),
    'Rf': Element(7, 4, 104, 'Rf', 261.0),
    'Db': Element(7, 5, 105, 'Db', 262.0),
    'Sg': Element(7, 6, 106, 'Sg', 266.0),
    'Bh': Element(7, 7, 107, 'Bh', 264.0),
    'Hs': Element(7, 8, 108, 'Hs', 269.0),
    'Mt': Element(7, 9, 109, 'Mt', 268.0),
    'Ds': Element(7, 10, 110, 'Ds', 269.0),
    'Rg': Element(7, 11, 111, 'Rg', 272.0),
    # Cn missing
    # Uut missing
    # Fl missing
    # Uup missing
    # Lv missing
    # Uus missing
    # Uuo missing
    # actinides
}

if __name__ == '__main__':
    import unittest
    from auxi.tools.chemistry.stoichiometry_test import StoichFunctionTester
    unittest.main()
